// SPDX-License-Identifier: AGPL-3.0-or-later
//! This module defines [HashMapWithDefault], which is a newtype wrapping a [HashMap] which allows
//! a default value to be defined.
//! When a key is not found in the [HashMap] the default value is returned, instead of [None].

use std::{borrow::Borrow, collections::HashMap, hash::Hash, ops::Index};

pub struct HashMapWithDefault<K, V> {
    hash_map: HashMap<K, V>,
    default: V,
}

impl<K, V> HashMapWithDefault<K, V> {
    pub fn new(default: V) -> Self {
        Self {
            hash_map: HashMap::new(),
            default,
        }
    }

    pub fn get_ref(&self) -> &HashMap<K, V> {
        &self.hash_map
    }

    pub fn get_mut(&mut self) -> &mut HashMap<K, V> {
        &mut self.hash_map
    }

    pub fn default(&self) -> &V {
        &self.default
    }

    pub fn mut_default(&mut self) -> &mut V {
        &mut self.default
    }
}

impl<K: Hash + Eq, V> HashMapWithDefault<K, V> {
    pub fn get<Q: ?Sized + Hash + Eq>(&self, index: &Q) -> &V
    where
        K: Borrow<Q>,
    {
        self.hash_map.get(index).unwrap_or(&self.default)
    }
}

impl<Q: ?Sized + Hash + Eq, K: Borrow<Q> + Hash + Eq, V> Index<&Q> for HashMapWithDefault<K, V> {
    type Output = V;
    fn index(&self, index: &Q) -> &Self::Output {
        self.get(index)
    }
}
